;; Generated package description from svg-lib.el  -*- no-byte-compile: t -*-
(define-package "svg-lib" "0.2.0.20211018.82904" "SVG tags, progress bars & icons" '((emacs "27.1")) :maintainer '("Nicolas P. Rougier" . "Nicolas.Rougier@inria.fr") :keywords '("svg" "icons" "tags" "convenience") :url "https://github.com/rougier/svg-lib")
